from dataclasses import dataclass
from datetime import datetime, timedelta

import pandas as pd

from .. import MarketDict
from .._typing import DemeterError
from ..broker import RowData


def to_minute(time: datetime) -> datetime:
    """
    second with 0 datetime
    :param time:
    :return:
    """
    return datetime(time.year, time.month, time.day, time.hour, time.minute)

"""
Note: in current version, all the market data should have the same index, which means the same timestamp range and
 interval, so we choose timestamp in default market to trigger actions. 
"""


class Trigger:
    def __init__(self, do, *args, **kwargs):
        self._do = do if do is not None else self.do_nothing
        self.kwargs = kwargs
        self.args = args

    def when(self, row_data: MarketDict[RowData]) -> bool:
        """
        when to handler data
        :param row_data: data in row
        :return:
        """
        return False

    def do_nothing(self, row_data: MarketDict[RowData], *args, **kwargs):
        pass

    def do(self, row_data: MarketDict[RowData]):
        """
        operation to handler with row data
        :param row_data:
        :return:
        """
        return self._do(row_data, *self.args, **self.kwargs)


class AtTimeTrigger(Trigger):
    """
    trigger action at a specific time
    """
    def __init__(self, time: datetime, do, *args, **kwargs):
        self._time = to_minute(time)
        super().__init__(do, *args, **kwargs)

    def when(self, row_data: MarketDict[RowData]) -> bool:
        return row_data.default.timestamp == self._time


class AtTimesTrigger(Trigger):
    """
    trigger action at some specific time
    """
    def __init__(self, time: [datetime], do, *args, **kwargs):
        self._time = [to_minute(t) for t in time]
        super().__init__(do, *args, **kwargs)

    def when(self, row_data: MarketDict[RowData]) -> bool:
        return self._time in row_data.default.timestamp


@dataclass
class TimeRange:
    start: datetime
    end: datetime


class TimeRangeTrigger(Trigger):
    """
    trigger action at a time range
    """
    def __init__(self, time_range: TimeRange, do, *args, **kwargs):
        self._time_range = TimeRange(to_minute(time_range.start), to_minute(time_range.end))
        super().__init__(do, *args, **kwargs)

    def when(self, row_data: MarketDict[RowData]) -> bool:
        return self._time_range.start <= row_data.default.timestamp < self._time_range.end


class TimeRangesTrigger(Trigger):
    """
    trigger action at some time range
    """
    def __init__(self, time_range: [TimeRange], do, *args, **kwargs):
        self._time_range: [TimeRange] = [TimeRange(to_minute(t.start), to_minute(t.end)) for t in time_range]
        super().__init__(do, *args, **kwargs)

    def when(self, row_data: MarketDict[RowData]) -> bool:
        for r in self._time_range:
            if r.start <= row_data.default.timestamp < r.end:
                return True
        return False


def check_time_delta(delta: timedelta):
    if delta.total_seconds() % 60 != 0:
        raise DemeterError("min time span is 1 minute")


class PeriodTrigger(Trigger):
    """
    trigger period action
    """
    def __init__(self, time_delta: timedelta, do, trigger_immediately=False, *args, **kwargs):
        self._next_match = None
        self._delta = time_delta
        self._trigger_immediately = trigger_immediately
        check_time_delta(time_delta)
        super().__init__(do, *args, **kwargs)

    def reset(self):
        self._next_match = None

    def when(self, row_data: MarketDict[RowData]) -> bool:
        if self._next_match is None:
            self._next_match = row_data.default.timestamp + self._delta
            return self._trigger_immediately

        if self._next_match == row_data.default.timestamp:
            self._next_match = self._next_match + self._delta
            return True

        return False


class PeriodsTrigger(Trigger):
    """
    trigger some period actions
    """
    def __init__(self, time_delta: [timedelta], do, trigger_immediately=False, *args, **kwargs):
        self._next_matches = [None for _ in time_delta]
        self._deltas = time_delta
        self._trigger_immediately = trigger_immediately

        for td in time_delta:
            check_time_delta(td)
        super().__init__(do, *args, **kwargs)

    def reset(self):
        self._next_matches = [None for _ in self._deltas]

    def when(self, row_data: MarketDict[RowData]) -> bool:
        if self._next_matches[0] is None:
            self._next_matches = [row_data.default.timestamp + d for d in self._deltas]
            return self._trigger_immediately

        for i in range(len(self._deltas)):
            if self._next_matches[i] == row_data.default.timestamp:
                self._next_matches[i] = self._next_matches[i] + self._deltas[i]
                return True

        return False



from typing import Dict, List

import pandas as pd

from .._typing import UnitDecimal, DemeterError, EvaluatorEnum
from ..broker import AccountStatus, AccountStatusCommon
from .math_helper import max_draw_down_fast, annualized_returns, get_benchmark_returns


class Evaluator(object):
    """
    calculate evaluator indicator for strategy.
    """

    def __init__(self, init_status: AccountStatus, data: pd.DataFrame | AccountStatusCommon, prices: pd.DataFrame):
        """
        init Evaluator
        :param init_status:
        :param data:
        :param prices:
        """
        self.init_status: AccountStatus = init_status
        self.end_status: AccountStatusCommon = data.iloc[-1]
        self.prices: pd.DataFrame = prices
        self.data: pd.DataFrame = data
        if len(data) < 2:
            raise DemeterError("not enough data")
        self.time_span_in_day = len(data.index) * (data.index[1] - data.index[0]).seconds / (60 * 60 * 24)
        self._result = None

    def run(self, enables: List[EvaluatorEnum]):
        """
        run evaluator
        :param enables:
        :return: result_dict
        """
        if EvaluatorEnum.ALL in enables:
            enables = [x for x in EvaluatorEnum]
            enables = filter(lambda x: x.value > 0, enables)
        result_dict: Dict[EvaluatorEnum, UnitDecimal] = {}
        for request in enables:
            match request:
                case EvaluatorEnum.ANNUALIZED_RETURNS:
                    result = UnitDecimal(annualized_returns(self.init_status.net_value,
                                                            self.end_status.net_value,
                                                            self.time_span_in_day), "")
                case EvaluatorEnum.BENCHMARK_RETURNS:
                    result = UnitDecimal(get_benchmark_returns(self.init_status.net_value,
                                                               self.prices.iloc[0],
                                                               self.prices.iloc[-1],
                                                               self.time_span_in_day), "")
                case EvaluatorEnum.MAX_DRAW_DOWN:
                    result = UnitDecimal(max_draw_down_fast(self.data.net_value), "")
                case _:
                    raise DemeterError(f"{request} has not implied")
            result_dict[request] = result
        self._result = result_dict
        return result_dict

    @property
    def result(self) -> Dict[EvaluatorEnum, UnitDecimal]:
        """
        return Evaluator._result property
        :return:
        """
        return self._result

    def __str__(self):
        """
        Evaluator print function
        :return:
        """
        str_array = []
        for k, v in self._result.items():
            str_array.append(f"{k.name}:{v}")
        return "; ".join(str_array)



import json
from decimal import Decimal
from types import SimpleNamespace
from functools import wraps

OUTPUT_WIDTH = 30


def to_decimal(value):
    """
    decimal value
    :param value:
    :return:
    """
    return Decimal(value)


def object_to_decimal(num):
    """
    convert object to decimal
    :param num:
    :return:
    """
    return Decimal(str(num)) if (isinstance(num, float) or type(num) == int) else num


def dict_to_object(dict_entity):
    """
    convert dict to object
    :param dict_entity:
    :return:
    """
    return json.loads(json.dumps(dict_entity), object_hook=lambda d: SimpleNamespace(**d))


def float_param_formatter(func):
    """
    decorator to convert param to float
    :param func:
    :return:
    """
    @wraps(func)
    def wrapper_func(*args, **kwargs):
        new_args = ()
        for arg in args:
            new_args += (object_to_decimal(arg),)
        for k, v in kwargs.items():
            kwargs[k] = object_to_decimal(v)
        return func(*new_args, **kwargs)

    return wrapper_func


def get_enum_by_name(me, name):
    """
    get enum by name
    :param me: enum
    :param name: key
    :return: get value in enum
    """
    for e in me:
        if e.name.lower() == name.lower():
            return e
    raise RuntimeError(f"cannot found {name} in {me}, allow value is " + str([x.name for x in me]))



from ._typing import RowData, BaseAction, MarketBalance, AccountStatus, MarketInfo, AccountStatusCommon, Asset, \
    MarketDict, AssetDict, ActionTypeEnum, MarketStatus, Rule
from .broker import Broker
from .market import Market



from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from typing import Dict

import pandas as pd

from .._typing import TokenInfo, DemeterError, UnitDecimal, PositionInfo
from ..broker import BaseAction, ActionTypeEnum, MarketBalance, MarketStatus
from ..utils import get_formatted_from_dict


class UniV3Pool(object):
    """
    pool information, corresponding with definition in pool contract.

    :param token0: First token in  pool contract.
    :type token0:  TokenInfo
    :param token1: Second token in  pool contract.
    :type token1: TokenInfo
    :param fee: fee rate of this pool, should be among [0.05, 0.3, 1]
    :type fee: float, 0.05
    :param base_token: which token will be considered as base token. eg: to a token pair of USDT/BTC, if you want price unit to be like 10000 usdt/btc, you should set usdt as base token, otherwise if price unit is 0.00001 btc/usdt, you should set btc as base token
    :type base_token: TokenInfo
    """

    def __init__(self, token0: TokenInfo, token1: TokenInfo, fee: float, base_token: TokenInfo):
        fee = Decimal(str(fee))
        self.token0 = token0
        self.token1 = token1
        self.is_token0_base = (base_token == token0)
        self.base_token = base_token
        self.tickSpacing = int(fee * 200)
        self.fee: Decimal = fee * Decimal(10000)
        self.fee_rate: Decimal = Decimal(fee) / Decimal(100)

    def __str__(self):
        """
        get string
        :return:
        :rtype:
        """
        return "PoolBaseInfo(Token0: {},".format(self.token0) + \
            "Token1: {},".format(self.token1) + \
            "fee: {},".format(self.fee_rate * Decimal(100)) + \
            "base token: {})".format(self.token0.name if self.is_token0_base else self.token1.name)


@dataclass
class UniLpBalance(MarketBalance):
    """
    current status of broker

    :param timestamp: timestamp
    :type timestamp: datetime
    :param base_uncollected: base token uncollect fee in all the positions.
    :type base_uncollected: UnitDecimal
    :param quote_uncollected: quote token uncollect fee in all the positions.
    :type quote_uncollected: UnitDecimal
    :param base_in_position: base token amount deposited in positions, calculated according to current price
    :type base_in_position: UnitDecimal
    :param quote_in_position: quote token amount deposited in positions, calculated according to current price
    :type quote_in_position: UnitDecimal
    :param pool_net_value: 按照池子base/quote关系的净值. 不是broker层面的(which 通常是对u的).
    :type pool_net_value: UnitDecimal
    :param price: current price
    :type price: UnitDecimal

    """
    base_uncollected: UnitDecimal
    quote_uncollected: UnitDecimal
    base_in_position: UnitDecimal
    quote_in_position: UnitDecimal
    position_count: int

    def get_output_str(self) -> str:
        """
        get colored and formatted string to output in console
        :return: formatted string
        :rtype: str
        """
        return get_formatted_from_dict({
            "total capital": self.pool_net_value.to_str(),
            "uncollect fee": f"{self.base_uncollected.to_str()},{self.quote_uncollected.to_str()}",
            "in position amount": f"{self.base_in_position.to_str()},{self.quote_in_position.to_str()}",
            "position count": self.position_count.to_str()
        })

    def to_array(self):
        return [
            self.base_uncollected,
            self.quote_uncollected,
            self.base_in_position,
            self.quote_in_position,
            self.position_count
        ]


@DeprecationWarning
class BrokerAsset(object):
    """
    Wallet of broker, manage balance of an asset.
    It will prevent excess usage on asset.
    """

    def __init__(self, token: TokenInfo, init_amount=Decimal(0)):
        self.token_info = token
        self.name = token.name
        self.decimal = token.decimal
        self.balance = init_amount

    def __str__(self):
        return f"{self.balance} {self.name}"

    def add(self, amount=Decimal(0)):
        """
        add amount to balance
        :param amount: amount to add
        :type amount: Decimal
        :return: entity itself
        :rtype: BrokerAsset
        """
        self.balance += amount
        return self

    def sub(self, amount=Decimal(0), allow_negative_balance=False):
        """
        subtract amount from balance. if balance is not enough, an error will be raised.

        :param amount: amount to subtract
        :type amount: Decimal
        :param allow_negative_balance: allow balance is negative
        :type allow_negative_balance: bool
        :return:
        :rtype:
        """
        base = self.balance if self.balance != Decimal(0) else Decimal(amount)

        if base == Decimal(0):  # amount and balance is both 0
            return self
        if allow_negative_balance:
            self.balance -= amount
        else:
            # if difference between amount and balance is below 0.01%, will deduct all the balance
            # That's because, the amount calculated by v3_core, has some acceptable error.
            if abs((self.balance - amount) / base) < 0.00001:
                self.balance = Decimal(0)
            elif self.balance - amount < Decimal(0):
                raise DemeterError(f"Insufficient balance, balance is {self.balance}{self.name}, "
                                   f"but sub amount is {amount}{self.name}")
            else:
                self.balance -= amount

        return self

    def amount_in_wei(self):
        return self.balance * Decimal(10 ** self.decimal)


@dataclass
class Position(object):
    """
    variables for position
    """

    pending_amount0: Decimal
    pending_amount1: Decimal
    liquidity: int


def position_dict_to_dataframe(positions: Dict[PositionInfo, Position]) -> pd.DataFrame:
    pos_dict = {
        "lower_tick": [],
        "upper_tick": [],
        "pending0": [],
        "pending1": [],
        "liquidity": []
    }
    for k, v in positions.items():
        pos_dict["lower_tick"].append(k.lower_tick)
        pos_dict["upper_tick"].append(k.upper_tick)
        pos_dict["pending0"].append(v.pending_amount0)
        pos_dict["pending1"].append(v.pending_amount1)
        pos_dict["liquidity"].append(v.liquidity)
    return pd.DataFrame(pos_dict)


@dataclass
class UniV3PoolStatus(MarketStatus):
    """
    current status of a pool, actuators can notify current status to broker by filling this entity
    """
    current_tick: int
    current_liquidity: int
    in_amount0: int
    in_amount1: int
    price: Decimal


@dataclass
class UniLpBaseAction(BaseAction):
    """
    Parent class of broker actions,

    :param base_balance_after: after action balance of base token
    :type base_balance_after: UnitDecimal
    :param quote_balance_after: after action balance of quote token
    :type quote_balance_after: UnitDecimal
    """

    base_balance_after: UnitDecimal
    quote_balance_after: UnitDecimal

    def get_output_str(self):
        return str(self)


@dataclass
class AddLiquidityAction(UniLpBaseAction):
    """
    Add Liquidity

    :param base_amount_max: inputted base token amount, also the max amount to deposit
    :type base_amount_max: ActionTypeEnum
    :param quote_amount_max: inputted base token amount, also the max amount to deposit
    :type quote_amount_max: datetime
    :param lower_quote_price: lower price base on quote token.
    :type lower_quote_price: UnitDecimal
    :param upper_quote_price: upper price base on quote token.
    :type upper_quote_price: UnitDecimal
    :param base_amount_actual: actual used base token
    :type base_amount_actual: UnitDecimal
    :param quote_amount_actual: actual used quote token
    :type quote_amount_actual: UnitDecimal
    :param position: generated position
    :type position: PositionInfo
    :param liquidity: liquidity added
    :type liquidity: int
    """
    base_amount_max: UnitDecimal
    quote_amount_max: UnitDecimal
    lower_quote_price: UnitDecimal
    upper_quote_price: UnitDecimal
    base_amount_actual: UnitDecimal
    quote_amount_actual: UnitDecimal
    position: PositionInfo
    liquidity: int

    def set_type(self):
        self.action_type = ActionTypeEnum.uni_lp_add_liquidity

    def get_output_str(self) -> str:
        """
        get colored and formatted string to output in console
        :return: formatted string
        :rtype: str
        """
        return f"""\033[1;31m{"Add liquidity":<20}\033[0m""" + \
            get_formatted_from_dict({
                "max amount": f"{self.base_amount_max.to_str()},{self.quote_amount_max.to_str()}",
                "price": f"{self.lower_quote_price.to_str()},{self.upper_quote_price.to_str()}",
                "position": self.position,
                "liquidity": self.liquidity,
                "balance": f"{self.base_balance_after.to_str()}(-{self.base_amount_actual.to_str()}), {self.quote_balance_after.to_str()}(-{self.quote_amount_actual.to_str()})"
            })


@dataclass
class CollectFeeAction(UniLpBaseAction):
    """
    collect fee

    :param position: position to operate
    :type position: PositionInfo
    :param base_amount: fee collected in base token
    :type base_amount: UnitDecimal
    :param quote_amount: fee collected in quote token
    :type quote_amount: UnitDecimal

    """
    position: PositionInfo
    base_amount: UnitDecimal
    quote_amount: UnitDecimal

    def set_type(self):
        self.action_type = ActionTypeEnum.uni_lp_collect

    def get_output_str(self) -> str:
        """
        get colored and formatted string to output in console
        :return: formatted string
        :rtype: str
        """
        return f"""\033[1;33m{"Collect fee":<20}\033[0m""" + \
            get_formatted_from_dict({
                "position": self.position,
                "balance": f"{self.base_balance_after.to_str()}(+{self.base_amount.to_str()}), {self.quote_balance_after.to_str()}(+{self.quote_amount.to_str()})"
            })


@dataclass
class RemoveLiquidityAction(UniLpBaseAction):
    """
    remove position

    :param position: position to operate
    :type position: PositionInfo
    :param base_amount: base token amount collected
    :type base_amount: UnitDecimal
    :param quote_amount: quote token amount collected
    :type quote_amount: UnitDecimal
    :param removed_liquidity: liquidity number has removed
    :type removed_liquidity: int
    :param remain_liquidity: liquidity number left in position
    :type remain_liquidity: int

    """
    position: PositionInfo
    base_amount: UnitDecimal
    quote_amount: UnitDecimal
    removed_liquidity: int
    remain_liquidity: int

    def set_type(self):
        self.action_type = ActionTypeEnum.uni_lp_remove_liquidity

    def get_output_str(self) -> str:
        """
        get colored and formatted string to output in console
        :return: formatted string
        :rtype: str
        """
        return f"""\033[1;32m{"Remove liquidity":<20}\033[0m""" + \
            get_formatted_from_dict({
                "position": self.position,
                "balance": f"{self.base_balance_after.to_str()}(+0), {self.quote_balance_after.to_str()}(+0)",
                "token_got": f"{self.base_amount.to_str()},{self.quote_amount.to_str()}",
                "removed liquidity": self.removed_liquidity,
                "remain liquidity": self.remain_liquidity
            })


@dataclass
class BuyAction(UniLpBaseAction):
    """
    buy token, swap from base token to quote token.

    :param amount: amount to buy(in quote token)
    :type amount: UnitDecimal
    :param price: price,
    :type price: UnitDecimal
    :param fee: fee paid (in base token)
    :type fee: UnitDecimal
    :param base_change: base token amount changed
    :type base_change: PositionInfo
    :param quote_change: quote token amount changed
    :type quote_change: UnitDecimal

    """
    amount: UnitDecimal
    price: UnitDecimal
    fee: UnitDecimal
    base_change: UnitDecimal
    quote_change: UnitDecimal

    def set_type(self):
        self.action_type = ActionTypeEnum.uni_lp_buy

    def get_output_str(self) -> str:
        """
        get colored and formatted string to output in console
        :return: formatted string
        :rtype: str
        """
        return f"""\033[1;36m{"Buy":<20}\033[0m""" + \
            get_formatted_from_dict({
                "price": self.price.to_str(),
                "fee": self.fee.to_str(),
                "balance": f"{self.base_balance_after.to_str()}(-{self.base_change.to_str()}), {self.quote_balance_after.to_str()}(+{self.quote_change.to_str()})"
            })


@dataclass
class SellAction(UniLpBaseAction):
    """
    sell token, swap from quote token to base token.

    :param amount: amount to sell(in quote token)
    :type amount: UnitDecimal
    :param price: price,
    :type price: UnitDecimal
    :param fee: fee paid (in quote token)
    :type fee: UnitDecimal
    :param base_change: base token amount changed
    :type base_change: PositionInfo
    :param quote_change: quote token amount changed
    :type quote_change: UnitDecimal

    """
    amount: UnitDecimal
    price: UnitDecimal
    fee: UnitDecimal
    base_change: UnitDecimal
    quote_change: UnitDecimal

    def set_type(self):
        self.action_type = ActionTypeEnum.uni_lp_sell

    def get_output_str(self):
        return f"""\033[1;37m{"Sell":<20}\033[0m""" + \
            get_formatted_from_dict({
                "price": self.price.to_str(),
                "fee": self.fee.to_str(),
                "balance": f"{self.base_balance_after.to_str()}(+{self.base_change.to_str()}), {self.quote_balance_after.to_str()}(-{self.quote_change.to_str()})"
            })



from decimal import Decimal
from enum import Enum
from typing import NamedTuple

DECIMAL_0 = Decimal(0)
DECIMAL_1 = Decimal(1)


class TimeUnitEnum(Enum):
    """
    time unit for moving average,

    * minute
    * hour
    * day
    """
    minute = 1
    hour = 60
    day = 60 * 24


class UnitDecimal(Decimal):
    """
    Decimal with unit, such a 1 eth.

    It's inherit from Decimal, but considering performance issues, calculate function is not override,
    so if you do calculate on this object, return type will be Decimal

    :param number: number to keep
    :type number: Decimal
    :param unit: unit of the number, eg: eth
    :type _unit: str
    :param output_format: output format, follow the document here: https://python-reference.readthedocs.io/en/latest/docs/functions/format.html
    :type output_format: str
    """
    __integral = Decimal(1)
    default_output_format = ".8g"

    def __new__(cls, value, unit: str = "", output_format=None):
        obj = Decimal.__new__(cls, value)
        obj._unit = unit
        obj.output_format: str = output_format if output_format is not None else UnitDecimal.default_output_format
        return obj

    def to_str(self):
        """
        get formatted string of this decimal. format is defined in self.output_format and unit will be append.

        :return: formatted string
        :rtype: str
        """
        dec = self.quantize(DECIMAL_1) if (self == self.to_integral() and self < 1e+29) else self.normalize()
        return "{:{}} {}".format(dec, self.output_format, self._unit)

    @property
    def unit(self):
        return self._unit

    @unit.setter
    def unit(self, value):
        self._unit = value


class EvaluatorEnum(Enum):
    ALL = 0
    ANNUALIZED_RETURNS = 1
    BENCHMARK_RETURNS = 2
    MAX_DRAW_DOWN = 3

    def __str__(self):
        return self.name


class TokenInfo(NamedTuple):
    """
    token info

    :param name: token symbol, will be set as unit of a token value, usdc
    :type name: str
    :param decimal: decimal of this token, 6
    :type decimal: int
    """
    name: str
    decimal: int


class PositionInfo(NamedTuple):
    """
    position information, including tick range and liquidity. It's the immute properties of a position

    :param lower_tick: lower tick
    :type lower_tick: int
    :param upper_tick: upper tick
    :type upper_tick: int

    """
    lower_tick: int
    upper_tick: int

    def __str__(self):
        return f"""tick:{self.lower_tick},{self.upper_tick}"""


class DemeterError(Exception):
    def __init__(self, message):
        self.message = message


class EthError(Exception):
    def __init__(self, code, message):
        self.code = code
        self.message = message



import math
from decimal import Decimal

from .liquitidy_math import get_sqrt_ratio_at_tick


def _x96_to_decimal(number: int):
    """
    decimal divide 2 ** 96
    :param number:
    :return:
    """
    return Decimal(number) / 2 ** 96


def decimal_to_x96(number: Decimal):
    """
    decimal multiple 2 ** 96
    :param number:
    :return:
    """
    return int(Decimal(number) * 2 ** 96)


def _x96_sqrt_to_decimal(sqrt_priceX96, token_decimal_diff=12):
    """
    sqrt_priceX96 divide token decimal
    :param sqrt_priceX96:
    :param token_decimal_diff:
    :return:
    """
    price = _x96_to_decimal(sqrt_priceX96)
    return (price ** 2) / 10 ** token_decimal_diff


## can round by spacing?
def sqrt_price_to_tick(sqrt_priceX96: int) -> int:
    """
    convert sqrt_priceX96 to tick data
    :param sqrt_priceX96:
    :return:
    """
    decimal_price = _x96_to_decimal(sqrt_priceX96)
    return pool_price_to_tick(decimal_price)


def pool_price_to_tick(price_decimal: Decimal):
    """
    pool price to tick data
    :param price_decimal:
    :return:
    """
    return int(math.log(price_decimal, math.sqrt(1.0001)))


def tick_to_sqrtPriceX96(tick: int):
    """
    convert tick data to sqrtPriceX96
    :param tick:
    :return:
    """
    return get_sqrt_ratio_at_tick(tick)


def tick_to_quote_price(tick: int, token_0_decimal, token_1_decimal, is_token0_base: bool):
    """
    tick data get quote price
    :param tick: tick data
    :param token_0_decimal: token0 decimal
    :param token_1_decimal: token1 decimal
    :param is_token0_base: base on token0
    :return: quote price
    """
    sqrt_price = get_sqrt_ratio_at_tick(tick)
    decimal_price = _x96_to_decimal(sqrt_price) ** 2
    pool_price = decimal_price * Decimal(10 ** (token_0_decimal - token_1_decimal))
    return Decimal(1 / pool_price) if is_token0_base else pool_price


def quote_price_to_tick(based_price: Decimal, token_0_decimal: int, token_1_decimal: int, is_token_base) -> int:
    """
    quote price to tick data
    :param based_price: base price
    :param token_0_decimal: token0 decimal
    :param token_1_decimal: token1 decimal
    :param is_token_base: base on token
    :return: tick data
    """
    # quote price->add decimal pool price->sqrt_price ->ticker
    sqrt_price = quote_price_to_sqrt(based_price, token_0_decimal, token_1_decimal, is_token_base)
    tick = sqrt_price_to_tick(sqrt_price)
    return tick


def quote_price_to_sqrt(based_price: Decimal, token_0_decimal: int, token_1_decimal: int, is_token_base) -> int:
    """
    convert quote price to sqrt
    :param based_price: price of base token
    :param token_0_decimal: token 0 decimal
    :param token_1_decimal: token 1 decimal
    :param is_token_base: if is base token
    :return: sqrt price
    """
    # quote price->add decimal pool price->sqrt_price ->ticker

    price = 1 / based_price if is_token_base else based_price
    pool_price = price / Decimal(10 ** (token_0_decimal - token_1_decimal))
    decimal_price = Decimal.sqrt(pool_price)
    return decimal_to_x96(decimal_price)


def from_wei(token_amt: int, decimal: int) -> Decimal:
    """
    token ** decimal
    :param token_amt:
    :param decimal:
    :return:
    """
    return Decimal(int(token_amt)) / Decimal(10 ** decimal)


def get_delta_gamma(lower_price: float, upper_price: float, price: float,
                    liquidity: int, decimal0: int, decimal1: int, is_0_base: bool):
    """
    get delta gamma
    :param lower_price: lower price
    :param upper_price: upper price
    :param price: price
    :param liquidity: liquidity
    :param decimal0: decimal 0
    :param decimal1: decimal 1
    :param is_0_base: check if token 0 is base
    :return:
    """
    lower_price_sqrtX96 = quote_price_to_sqrt(Decimal(lower_price), decimal0, decimal1, is_0_base)
    upper_price_sqrtX96 = quote_price_to_sqrt(Decimal(upper_price), decimal0, decimal1, is_0_base)
    if lower_price_sqrtX96 > upper_price_sqrtX96:
        (lower_price_sqrtX96, upper_price_sqrtX96) = (upper_price_sqrtX96, lower_price_sqrtX96)
    return get_delta_gamma_sqrtX96(lower_price,
                                   lower_price_sqrtX96,
                                   upper_price,
                                   upper_price_sqrtX96,
                                   price,
                                   liquidity,
                                   decimal0, decimal1,
                                   is_0_base)


def get_delta_gamma_sqrtX96(lower_price, sqrtA: int,
                            upper_price, sqrtB: int,
                            price,
                            liquidity: int,
                            d0: int,
                            d1: int,
                            is_0_base: bool):
    """
    k = 2 ** 96
    a0 = k * (10**(-d)) * Liquidity * (1/SqrtPrice - 1/upper_price_sqrtX96)
    a1= Liquidity / k / 10**d * (SqrtPrice - lower_price_sqrtX96)


    if 0 base:
    SqrtPrice=k / (10 ** (d/2)) / (p**0.5)
    net_value = a1 * p                    price <= lower, a0 is constant
                a0 + a1 * p               lower < price < upper
                a0                        price >= upper

    a0 + a1 * p = liquidity * 10 ** (0.5 * d) / 10 ** d0 * price_float ** 0.5 - \
              k / upper_sqrt * liquidity / 10 ** d0 + \
              liquidity*  price_float ** 0.5 / 10 ** d1 / 10 ** (0.5 * d) - \
              lower_sqrt / k * price_float * liquidity / 10 ** d1


    if 1 base
    SqrtPrice = k * p**0.5 / (10 ** (d/2))

    net_value = a0 * p                         price <= lower, a0 is constant
                a0 * p + a1                    lower < price < upper
                a1                             price >= upper, a1 is constant

    a0 * p + a1 = liquidity * price_float ** 0.5 * 10 ** (0.5 * d) / 10 ** d0 - \
                  k / upper_sqrt * liquidity * price_float / 10 ** d0 + \
                  liquidity * price_float ** 0.5 / 10 ** d1 / 10 ** (0.5 * d) - \
                  lower_sqrt / k * liquidity / 10 ** d1



    a0 + p * a1 =  Liquidity / 10**(d/2) / p**(1/2) + Liquidity * p**(1.5) / 10 ** (1.5*d) -
                   Liquidity * lower_price_sqrtX96 * p / 2**96 / 10**d
    """
    k = 2 ** 96
    d = d0 - d1
    if is_0_base:
        if price <= lower_price:
            delta = liquidity / 2 ** 96 / 10 ** d1 * (sqrtB - sqrtA)
            gamma = 0
        elif lower_price < price < upper_price:
            m = 10 ** (0.5 * d)
            delta = liquidity * (0.5 * m / price ** 0.5 / 10 ** d0 + \
                                 0.5 / 10 ** d1 / m / price ** 0.5 - \
                                 sqrtA / k / 10 ** d1)
            gamma = -0.25 * liquidity / price ** 1.5 * (m / 10 ** d0 + 1 / 10 ** d1 / m)
        else:
            delta = 0
            gamma = 0
    else:
        if price <= lower_price:
            delta = liquidity / 10 ** d0 * (k / sqrtA - k / sqrtB)
            gamma = 0
        elif lower_price < price < upper_price:
            m = 10 ** (0.5 * d)
            delta = liquidity * (0.5 * m / price ** 0.5 / 10 ** d0 + \
                                 0.5 / 10 ** d1 / m / price ** 0.5 - \
                                 k / sqrtB / 10 ** d0)
            gamma = -0.25 * liquidity / price ** 1.5 * (m / 10 ** d0 + 1 / m / 10 ** d1)
        else:
            delta = 0
            gamma = 0

    return delta, gamma



"""
uniswap market
"""

from .market import UniLpMarket
from ._typing import UniV3Pool, UniV3PoolStatus, BrokerAsset, Position, UniLpBalance
from .core import V3CoreLib
from .data import LineTypeEnum, UniLPData



# from enum import Enum
# from typing import NamedTuple, Dict
#
# import pandas as pd
# from pandas import _typing as pd_typing
#
# from ._typing import DemeterError
#
# DEFAULT_AGG_METHOD = "first"
#
#
# class Rule(NamedTuple):
#     agg: str | None
#     fillna_method: str | None
#     fillna_value: int | None
#
#
# EMPTY_RULE = Rule(None, None, None)
#
#
# def get_line_rules_safe(key: str) -> Rule:
#     if key in LINE_RULES:
#         return LINE_RULES[key]
#     else:
#         return EMPTY_RULE
#
#
# class LineTypeEnum(Enum):
#     """
#     predefined column, used to define fillna method.
#     """
#     timestamp = 1
#     netAmount0 = 2
#     netAmount1 = 3
#     closeTick = 4
#     openTick = 5
#     lowestTick = 6
#     highestTick = 7
#     inAmount0 = 8
#     inAmount1 = 9
#     currentLiquidity = 10
#     other = 100
#
#
# LINE_RULES = {
#     LineTypeEnum.timestamp.name: EMPTY_RULE,
#     LineTypeEnum.netAmount0.name: Rule("sum", None, 0),
#     LineTypeEnum.netAmount1.netAmount0: Rule("sum", None, 0),
#     LineTypeEnum.closeTick.name: Rule("last", "ffill", None),
#     LineTypeEnum.openTick.name: Rule("first", "ffill", None),
#     LineTypeEnum.lowestTick.name: Rule("min", "ffill", None),
#     LineTypeEnum.highestTick.name: Rule("max", "ffill", None),
#     LineTypeEnum.inAmount0.name: Rule("sum", None, 0),
#     LineTypeEnum.inAmount1.name: Rule("sum", None, 0),
#     LineTypeEnum.currentLiquidity.name: Rule("sum", "ffill", None),
# }
#
#
# class Cursorable(object):
#     """
#     give Lines or line a cursor. which help to access current row with index 0 in a loop.  eg:
#
#     if data= [1,2,3,4,5], self.cursor=0
#
#     * data.get_by_cursor[0]==1
#     * data.get_by_cursor[2]==3
#     * data.get_by_cursor[4]==5
#
#     if self.cursor=2
#
#     * data.get_by_cursor[-2]==1, access previse row with negative index,
#     * data.get_by_cursor[0]==3, access current row with 0,
#     * data.get_by_cursor[2]==5, access after row with positive index
#
#     if self.cursor=4
#
#     * data.get_by_cursor[-4]==1
#     * data.get_by_cursor[-2]==3
#     * data.get_by_cursor[0]==5
#     """
#
#     def __init__(self):
#         self.cursor = 0
#
#     def row_length(self):
#         """
#         get total length
#
#         """
#         return self.index.size
#
#     def get_cursor_range(self):
#         """
#         get legal range for index
#         :return: min, max
#         """
#         return -self.cursor, self.index.size - self.cursor - 1
#
#     def move_cursor_to_next(self):
#         """
#         move cursor to on_bar index
#         """
#         self.cursor += 1
#
#     def reset_cursor(self):
#         """
#         reset cursor to zero
#         """
#         self.cursor = 0
#
#     def get_by_cursor(self, i):
#         """
#          access row by cursor
#
#          :param i: index
#          :type i: int
#          :return: selected row, in form of pd.Series
#          """
#         real_index = self.cursor + i
#
#         if real_index < 0 or real_index >= self.index.size:
#             raise IndexError("index out of range")
#         return self.iloc[real_index]
#
#
# class Line(pd.Series, Cursorable):
#     """
#     a column of data, inherit from pandas.Series
#
#     and it has two additional function:
#
#     * it has a property of column type. if the column type is predefined, then fillna and resample method is not changeable
#     * it provide cursor.
#
#     note: index of this Series must be datetime
#     """
#
#     def __init__(
#             self,
#             data=None,
#             index=None,
#             dtype: pd_typing.Dtype | None = None,
#             name=None,
#             copy: bool = False,
#             fastpath: bool = False
#     ):
#
#         if not isinstance(index, pd.core.indexes.datetimes.DatetimeIndex):
#             raise DemeterError("index must be datetime")
#         super().__init__(data, index, dtype, name, copy, fastpath)
#         Cursorable.__init__(self)
#
#     def resample_by_type(
#             self,
#             rule,
#             axis=0,
#             closed: str | None = None,
#             label: str | None = None,
#             convention: str = "start",
#             kind: str | None = None,
#             loffset=None,
#             base: int | None = None,
#             on=None,
#             level=None,
#             origin: str | pd_typing.TimestampConvertibleTypes = "start_day",
#             offset: pd_typing.TimedeltaConvertibleTypes | None = None,
#             agg=DEFAULT_AGG_METHOD
#     ) -> pd.Series:
#         """
#         with this method, predefined type will be resampled correctly, other param is the same to Series.resample()
#
#         :param agg: if line_type is other, should define resample method, such as sum, mean
#         :return: resampled data
#         """
#         resampler = super().resample(rule, axis, closed, label, convention, kind, loffset, base, on, level, origin,
#                                      offset)
#         rule = get_line_rules_safe(self.name)
#         agg_method = rule.agg
#         if agg_method is None:
#             agg_method = agg
#         return resampler.agg(agg_method)
#
#     def fillna(
#             self,
#             value: object | pd_typing.ArrayLike | None = None,
#             method: pd_typing.FillnaOptions | None = None,
#             axis=None,
#             inplace=False,
#             limit=None,
#             downcast=None,
#     ) -> pd.Series:
#         """
#         fill empty item. param is the same to pandas.Series.fillna
#
#         if line type is predefined, method and value will be omitted, and data will be filled as predefined
#
#         """
#         param = get_line_rules_safe(self.name)
#         if param.fillna_method:
#             method = param.fillna_method
#         if param.fillna_value is not None:
#             value = param.fillna_value
#         return super().fillna(
#             value=value,
#             method=method,
#             axis=axis,
#             inplace=inplace,
#             limit=limit,
#             downcast=downcast,
#         )
#
#     @staticmethod
#     def from_series(series: pd.Series, row_index: 0) -> "Line":
#         """
#         convert serise to line, line type will be defined by series.name
#
#         :param series: series to process
#         :type series: pd.Series
#         :param row_index: new row index
#         :type row_index: int
#         :return: converted Line
#         :rtype: Line
#         """
#         line = series.copy(deep=True)
#         line.cursor = row_index
#         return line
#
#
# class Lines(pd.DataFrame, Cursorable):
#     """
#     a data table, inherit from pandas.DataFrame
#
#     and it has two additional function:
#
#     * it has a property of column type. if the column type is predefined, then fillna and resample method can not change
#     * it provide cursor. assistant to access current row in a loop
#
#     note: index of this DataFrame must be datetime
#     """
#
#     def __init__(
#             self,
#             data=None,
#             index: pd_typing.Axes | None = None,
#             columns: pd_typing.Axes | None = None,
#             dtype: pd_typing.Dtype | None = None,
#             copy: bool | None = None,
#     ):
#         if (index is not None) and (not isinstance(index, pd.core.indexes.datetimes.DatetimeIndex)):
#             raise DemeterError("index must be datetime")
#         # add line by type
#         line_data = dict()
#         if data is not None:
#             if isinstance(data, list):
#                 for item in data:
#                     Lines.__set_line_name_by_object(item, line_data)
#             # after line is added, it's type will become series, so we identify its type by name
#             elif isinstance(data, dict):
#                 line_data = data
#             else:
#                 Lines.__set_line_name_by_object(data, line_data)
#         super().__init__(line_data, index, columns, dtype, copy)
#         Cursorable.__init__(self)
#
#     @staticmethod
#     def __set_line_name_by_object(item, line_data):
#         """
#         after line is added, it's type will become series, so we set their type to name.
#         """
#         if isinstance(item, pd.Series) and item.name:
#             line_data[item.name] = item
#         else:
#             line_data["Column" + str(len(line_data.keys()))] = item
#
#     def append_column(self, name: str, column: Line):
#         """
#         add a column to Lines
#
#         :param name: column name
#         :type name: str
#         :param column: column to add, type must be line, or pandas.Series, but index of Series should be datetime
#         :type: Line
#
#         """
#         if not isinstance(column.index, pd.core.indexes.datetimes.DatetimeIndex):
#             raise DemeterError("index must be datetime")
#         if self.data.columns.size == 0:
#             self.data.index = column.index
#         self.data[name] = column
#
#     def fillna(
#             self,
#             value: object | pd_typing.ArrayLike | None = None,
#             method: pd_typing.FillnaOptions | None = None,
#             axis: pd_typing.Axis | None = None,
#             inplace: bool = False,
#             limit=None,
#             downcast=None,
#     ) -> pd.DataFrame | None:
#         """
#         fill empty item. param is the same to pandas.Series.fillna
#
#         if column name is predefined, method and value will be omitted, and data will be filled as predefined
#
#         """
#         new_df = self.copy(False)
#
#         # fill close tick first, it will be used later.
#         if LineTypeEnum.closeTick.name in new_df.columns:
#             new_df[LineTypeEnum.closeTick.name] = \
#                 new_df[LineTypeEnum.closeTick.name].fillna(value=None,
#                                                            method=get_line_rules_safe(
#                                                                LineTypeEnum.closeTick.name).fillna_method,
#                                                            axis=axis,
#                                                            inplace=inplace,
#                                                            limit=limit,
#                                                            downcast=downcast)
#         for column_name in new_df.columns:
#             if column_name == LineTypeEnum.closeTick.name:
#                 continue
#             rule = get_line_rules_safe(column_name)
#             if not rule.fillna_method and rule.fillna_value is not None:
#                 new_df[column_name] = new_df[column_name].fillna(value, method, axis, inplace, limit, downcast)
#             else:
#                 current_method = rule.fillna_method if rule.fillna_method else method
#                 current_value = rule.fillna_value if rule.fillna_value else value
#                 # all tick related field will be filled with close_tick.
#                 if column_name in [LineTypeEnum.openTick.name,
#                                    LineTypeEnum.highestTick.name,
#                                    LineTypeEnum.lowestTick.name] and LineTypeEnum.closeTick.name in new_df.columns:
#                     current_method = None
#                     current_value = new_df[LineTypeEnum.closeTick.name]
#                 new_df[column_name] = new_df[column_name].fillna(value=current_value, method=current_method, axis=axis,
#                                                                  inplace=inplace,
#                                                                  limit=limit,
#                                                                  downcast=downcast)
#         new_lines = Lines.from_dataframe(new_df)
#         new_lines.cursor = self.cursor
#         return new_lines
#
#     def get_line(self, index: int = None, name: str = None) -> Line:
#         """
#         get a column of Lines
#
#         :param index: column index, if index is set, name param will be override
#         :type index: int
#         :param name: column name
#         :type name: str
#         :return: selected column
#         :rtype: Line
#         """
#         if not index and not name:
#             raise DemeterError("must use index or name")
#         if index:
#             return Line.from_series(self.iloc[:, index], self.cursor)
#         elif name:
#             return Line.from_series(self[name], self.cursor)
#
#     def resample_by_type(
#             self,
#             rule,
#             axis=0,
#             closed: str | None = None,
#             label: str | None = None,
#             convention: str = "start",
#             kind: str | None = None,
#             loffset=None,
#             base: int | None = None,
#             on=None,
#             level=None,
#             origin: str | pd_typing.TimestampConvertibleTypes = "start_day",
#             offset: pd_typing.TimedeltaConvertibleTypes | None = None,
#             agg: Dict[str, str] = {}
#     ) -> "Lines":
#         """
#         with this method, predefined column will be resampled correctly, other param is the same to Series.resample()
#
#         :param agg: if line name is not predefined, should define resample method, such as sum, mean
#         :return: resampled data
#         """
#         resampler = super().resample(rule, axis, closed, label, convention, kind, loffset, base, on, level, origin,
#                                      offset)
#         agg_dict = {}
#         for column_name in self.columns:
#             rule = get_line_rules_safe(column_name)
#             agg_method = rule.agg
#             if agg_method is None:
#                 if column_name in agg:
#                     agg_method = agg[column_name]
#                 else:
#                     agg_method = DEFAULT_AGG_METHOD
#             agg_dict[column_name] = agg_method
#         df = resampler.agg(agg_dict)
#         new_lines = Lines.from_dataframe(df)
#         new_lines.cursor = 0
#         return new_lines
#
#     @staticmethod
#     def from_dataframe(df: pd.DataFrame) -> "Lines":
#         """
#         convert pandas.DataFrame to Lines. this method will copy inner variable to new Lines.
#
#         :param df: dataframe to convert
#         :type df: DataFrame
#         :return: Lines
#         :rtype: Lines
#         """
#         new_lines = Lines(index=df.index)
#         new_lines._mgr = df._mgr
#         return new_lines
#
#     @staticmethod
#     def load_downloaded(path: str) -> "Lines":
#         """
#         load data which saved by downloader as lines
#         :param path: csv file path
#         :type path: str
#         :return: loaded data
#         :rtype: Lines
#         """
#         df = pd.read_csv(path)
#         df["timestamp"] = pd.to_datetime(df["timestamp"])
#         df.set_index("timestamp", inplace=True)
#         return Lines.from_dataframe(df)



from .application import float_param_formatter, to_decimal
from .console_text import ForColorEnum, BackColorEnum, ModeEnum, get_formatted, STYLE, \
    get_formatted_predefined, get_formatted_from_dict



from decimal import Decimal
from typing import Dict

import pandas as pd

from ._typing import Asset, TokenInfo, AccountStatus, MarketDict, AssetDict
from .market import Market
from .._typing import DemeterError, UnitDecimal
from ..utils import get_formatted_from_dict, get_formatted_predefined, STYLE, \
    float_param_formatter


class Broker:
    def __init__(self, allow_negative_balance=False, record_action_callback=None):
        """
        init Broker
        :param allow_negative_balance:
        :param record_action_callback:
        """
        self.allow_negative_balance = allow_negative_balance
        self._assets: AssetDict[Asset] = AssetDict()
        self._markets: MarketDict[Market] = MarketDict()
        self._record_action_callback = record_action_callback

    # region properties

    @property
    def markets(self) -> MarketDict[Market]:
        """
        markets property
        :return:
        """
        return self._markets

    @property
    def assets(self) -> AssetDict[Asset]:
        """
        asset property
        :return:
        """
        return self._assets

    # endregion

    def __str__(self):
        """
        Broker raw info
        :return:
        """
        return "assets: " + ",".join([f"({v})" for k, v in self._assets.items()]) + \
            "; markets: " + ",".join([f"({v})" for k, v in self.markets.items()])

    def add_market(self, market: Market):
        """
        Set a new market to broker,
        User should initialize market before set to broker(Because there are too many initial parameters)
        :param market: market
        :type market: Market
        :return:
        :rtype:
        """
        if market.market_info in self._markets:
            raise DemeterError("market has exist")
        self._markets[market.market_info] = market
        market.broker = self
        market._record_action_callback = self._record_action_callback

    @float_param_formatter
    def add_to_balance(self, token: TokenInfo, amount: Decimal | float):
        """
        set initial balance for token

        :param token: which token to set
        :type token: TokenInfo
        :param amount: balance, eg: 1.2345
        :type amount: Decimal | float
        """
        if token in self._assets:
            asset: Asset = self._assets[token]
        else:
            asset: Asset = self.__add_asset(token)
        asset.add(amount)
        return asset

    @float_param_formatter
    def set_balance(self, token: TokenInfo, amount: Decimal | float):
        """
        set the balance value
        :param token: the token of asset, TokenInfo(name='usdc', decimal=6)
        :param amount: amount of token, 10000
        :return: Asset instance, usdc: 10000
        """
        asset: Asset = self.__add_asset(token)
        asset.balance = amount
        return asset

    @float_param_formatter
    def subtract_from_balance(self, token: TokenInfo, amount: Decimal | float):
        """

        :param token: TokenInfo object
        :param amount: Decimal or float type
        :return:
        """
        if token in self._assets:
            asset: Asset = self._assets[token]
            asset.sub(amount, allow_negative_balance=self.allow_negative_balance)
        else:
            if self.allow_negative_balance:
                asset: Asset = self.__add_asset(token)
                asset.balance = 0 - amount
            else:
                raise DemeterError(f"{token.name} doesn't exist in assets dict")
        return asset

    def __add_asset(self, token: TokenInfo) -> Asset:
        """
        set Asset with Token
        :param token: TokenInfo
        :return: Asset with TokenInfo
        """
        self._assets[token] = Asset(token, 0)
        return self._assets[token]

    def get_token_balance(self, token: TokenInfo):
        """
        get balance of token
        :param token:
        :return:
        """
        if token in self.assets:
            return self._assets[token].balance
        else:
            raise DemeterError(f"{token.name} doesn't exist in assets dict")

    def get_token_balance_with_unit(self, token: TokenInfo):
        """
        get balance of token with unit
        :param token: TokenInfo
        :return: UnitDecimal balance
        """
        return UnitDecimal(self.get_token_balance(token), token.name)

    def get_account_status(self, prices: pd.Series | Dict[str, Decimal], timestamp=None) -> AccountStatus:
        """
        get account status
        :param prices: price series
        ('eth', Decimal('1610.553895752868641174609110')) ('usdc', 1)
        :param timestamp: optional timestamp, 2022-08-20 00:00:00
        :return: account status
        AccountStatus(timestamp=None, net_value=Decimal('26105.53895752868641174609110'), asset_balances=<demeter.broker._typing.AssetDict object at 0x11842f7c0>, market_status=<demeter.broker._typing.MarketDict object at 0x11842e020>)
        """
        account_status = AccountStatus(timestamp=timestamp)
        for k, v in self.markets.items():
            account_status.market_status[k] = v.get_market_balance(prices)
        account_status.market_status.set_default_key(self.markets.get_default_key())
        for k, v in self.assets.items():
            account_status.asset_balances[k] = v.balance
        asset_sum = sum([v * prices[k.name] for k, v in account_status.asset_balances.items()])
        market_sum = sum([v.net_value for v in account_status.market_status.values()])
        account_status.net_value = asset_sum + market_sum
        return account_status

    def formatted_str(self):
        """
        get formatted broker info
        :return: str info of broker
        """
        str_to_print = get_formatted_predefined("Broker", STYLE["header1"]) + "\n"

        str_to_print += get_formatted_predefined("Asset amounts", STYLE["header2"]) + "\n"
        balances = {}
        for asset in self._assets.values():
            balances[asset.name] = asset.balance
        str_to_print += get_formatted_from_dict(balances) + "\n"
        str_to_print += get_formatted_predefined("Markets", STYLE["header2"]) + "\n"
        for market in self._markets.values():
            str_to_print += market.formatted_str() + "\n"
        return str_to_print



import logging
from datetime import datetime
from decimal import Decimal
from typing import Dict

import pandas as pd

from ._typing import BaseAction, MarketBalance, MarketStatus, MarketInfo
from .._typing import DECIMAL_0, DemeterError

DEFAULT_DATA_PATH = "./data"


class Market:
    """
    note: only get properties are allow in this base class
    """

    def __init__(self,
                 market_info: MarketInfo,
                 data: pd.DataFrame = None,
                 data_path=DEFAULT_DATA_PATH):
        """
        init Market
        :param market_info: uni_market
        :param data: None or dataframe data
        :param data_path: default ./data dir
        """
        self._data: pd.DataFrame = data
        self._market_info: MarketInfo = market_info
        self.broker = None
        self._record_action_callback = lambda x: x  # default value, will be set by broker
        self.data_path: str = data_path
        self.logger = logging.getLogger(__name__)
        self._market_status = MarketStatus(None)

    def __str__(self):
        return f"{self._market_info.name}:{type(self).__name__}"

    @property
    def market_info(self) -> MarketInfo:
        return self._market_info

    @property
    def data(self):
        """
        data got from uniswap pool
        :return:
        :rtype:
        """
        return self._data

    @data.setter
    def data(self, value):
        if isinstance(value, pd.DataFrame):
            self._data = value
        else:
            raise ValueError()

    def record_action(self, action: BaseAction):
        if self._record_action_callback is not None:
            self._record_action_callback(action)

    # region for subclass to override
    def check_asset(self):
        pass

    def update(self):
        """
        update status various in markets. eg. liquidity fees of uniswap
        :return:
        :rtype:
        """
        pass

    @property
    def market_status(self):
        return self._market_status

    def set_market_status(self, timestamp: datetime, data: pd.Series | MarketStatus):
        """
        set up market status, such as liquidity, price
        :param timestamp: current timestamp
        :type timestamp: datetime
        :param data: market status
        :type data: pd.Series | MarketStatus
        """
        if isinstance(data, MarketStatus):
            self._market_status = data
        else:
            self._market_status = MarketStatus(timestamp)

    def get_market_balance(self, prices: pd.Series | Dict[str, Decimal]) -> MarketBalance:
        """
        get market asset balance
        :param prices: current price of each token
        :type prices: pd.Series | Dict[str, Decimal]
        :return:
        :rtype:
        """
        return MarketBalance(DECIMAL_0)

    def check_before_test(self):
        """
        do some check for this market before back test start
        :return:
        :rtype:
        """
        if not isinstance(self.data, pd.DataFrame):
            raise DemeterError("data must be type of data frame")
        if not isinstance(self.data.index, pd.core.indexes.datetimes.DatetimeIndex):
            raise DemeterError("date index must be datetime")

    def formatted_str(self):
        return ""

    # endregion



from enum import Enum
from typing import Dict


class ForColorEnum(Enum):
    default = 0
    black = 30
    red = 31
    green = 32
    yellow = 33
    blue = 34
    purple = 35
    cyan = 36
    white = 37


class BackColorEnum(Enum):
    default = 0
    black = 40
    red = 41
    green = 42
    yellow = 43
    blue = 44
    purple = 45
    cyan = 46
    white = 47


class ModeEnum(Enum):
    normal = 0
    bold = 1
    underline = 4
    blink = 5
    invert = 7
    hide = 8


DEFAULT_END = 0

STYLE = {
    "header1": {
        "mode": ModeEnum.invert,
        "fore": ForColorEnum.red,
        "back": BackColorEnum.default,
        "width": 50
    },
    "header2": {
        "mode": ModeEnum.invert,
        "fore": ForColorEnum.purple,
        "back": BackColorEnum.default,
        "width": 30
    },
    "header3": {
        "mode": ModeEnum.underline,
        "fore": ForColorEnum.yellow,
        "back": BackColorEnum.default,
        "width": -1
    },
    "key": {
        "mode": ModeEnum.normal,
        "fore": ForColorEnum.blue,
        "back": BackColorEnum.default,
        "width": 10
    },
    "value": {
        "mode": ModeEnum.normal,
        "fore": ForColorEnum.default,
        "back": BackColorEnum.default,
        "width": 25
    }
}


def get_formatted(string: str,
                  mode: ModeEnum = ModeEnum.normal,
                  fore: ForColorEnum = ForColorEnum.default,
                  back: BackColorEnum = BackColorEnum.default,
                  width=-1) -> str:
    mode = '{}'.format(mode.value if mode != mode.normal else '')
    fore = '{}'.format(fore.value if fore != ForColorEnum.default else "")
    back = '{}'.format(back.value if back != BackColorEnum.default else "")
    style = ';'.join([s for s in [mode, fore, back] if s])
    end = ""
    if style != "":
        style = """\033[{}m""".format(style)
        end = """\033[0m"""
    if width > 0:
        string = '{}{:<{}}{}'.format(style, string, width, end)
    else:
        string = '{}{}{}'.format(style, string, end)

    return string


def get_formatted_predefined(string: str, style):
    return get_formatted(string, style["mode"], style["fore"], style["back"], style["width"])


def get_formatted_from_dict(values: Dict[str, str]) -> str:
    str_array = []
    for k, v in values.items():
        str_array.append(
            f"{get_formatted_predefined(k, STYLE['key'])}:{get_formatted_predefined(str(v), STYLE['value'])}")
    return "".join(str_array)


# def TestColor():
#     print(get_formatted('正常显示'))
#     print('')
#     print("测试显示模式")
#     print(get_formatted('高亮', mode=ModeEnum.bold), )
#     print(get_formatted('下划线', mode=ModeEnum.underline), )
#     print(get_formatted('闪烁', mode=ModeEnum.blink), )
#     print(get_formatted('反白', mode=ModeEnum.invert), )
#     print(get_formatted('不可见', mode=ModeEnum.hide))
#     print('')
#     print("测试前景色")
#     print(get_formatted('黑色', fore=ForColorEnum.black), )
#     print(get_formatted('红色', fore=ForColorEnum.red), )
#     print(get_formatted('绿色', fore=ForColorEnum.green), )
#     print(get_formatted('黄色', fore=ForColorEnum.yellow), )
#     print(get_formatted('蓝色', fore=ForColorEnum.blue), )
#     print(get_formatted('紫红色', fore=ForColorEnum.purple), )
#     print(get_formatted('青蓝色', fore=ForColorEnum.cyan), )
#     print(get_formatted('白色', fore=ForColorEnum.white))
#     print('')
#     print("测试背景色")
#     print(get_formatted('黑色', back=BackColorEnum.black), )
#     print(get_formatted('红色', back=BackColorEnum.red), )
#     print(get_formatted('绿色', back=BackColorEnum.green), )
#     print(get_formatted('黄色', back=BackColorEnum.yellow), )
#     print(get_formatted('蓝色', back=BackColorEnum.blue), )
#     print(get_formatted('紫红色', back=BackColorEnum.purple), )
#     print(get_formatted('青蓝色', back=BackColorEnum.cyan), )
#     print(get_formatted('白色', back=BackColorEnum.white))
#     print('')
#     print(get_formatted("综合", ModeEnum.invert, ForColorEnum.red))
#
#
# if __name__ == '__main__':
#     TestColor()



from typing import Dict, List

import pandas as pd

from .trigger import Trigger
from .. import Broker, MarketDict, AccountStatus, AssetDict, Asset
from .._typing import DemeterError
from ..broker import MarketInfo, RowData, BaseAction, Market


class Strategy(object):
    """
    strategy parent class, all user strategy should inherit this class
    """

    def __init__(self):
        self.broker: Broker = None
        self.data: MarketDict[pd.DataFrame] = MarketDict()
        self.markets: MarketDict[Market] = MarketDict()
        self.number_format = ".8g"
        self.prices: pd.DataFrame = None
        self.triggers: [Trigger] = []
        self.account_status: List[AccountStatus] = []
        self.assets: AssetDict[Asset] = AssetDict()
        self.actions: List[BaseAction] = []

    def initialize(self):
        """
        initialize your strategy, this will be called before self.on_bar()

        """
        pass

    def before_bar(self, row_data: MarketDict[RowData]):
        """
        called before triggers on each row, at this time, fees are not updated yet. you can add some indicator or add some actions

        :param row_data: row data, include columns load from data, converted data( price, volumn, and timestamp, index), indicators(such as ma)
        :type row_data: Union[{MarketInfo:RowData}, pd.Series]

        """

        pass

    def on_bar(self, row_data: MarketDict[RowData]):
        """
        called after triggers on each row, at this time, fees and account status are not updated yet. you can add some actions here

        :param row_data: row data, include columns load from data, converted data( price, volumn, and timestamp, index), indicators(such as ma)
        :type row_data: Union[{MarketInfo:RowData}, pd.Series]
        """
        pass

    def after_bar(self, row_data: MarketDict[RowData]):
        """
        called after fees and account status are updated on each row. you can add some statistic logic here

        :param row_data: row data, include columns load from data, converted data( price, volumn, and timestamp, index), indicators(such as ma)
        :type row_data: Union[{MarketInfo:RowData}, pd.Series]
        """
        pass

    def finalize(self):
        """
        this will run after all the data processed.

        """
        pass

    def notify(self, action: BaseAction):
        """
        notify if non-basic action happens

        :param action:  action
        :type action: BaseAction
        """
        pass

    def _add_column(self, market: MarketInfo | Market, name: str, line: pd.Series):
        """
        add a column to data

        :param name: column name, sma
        :type name: str
        :param market: market1
        :type market: MarketInfo
        :param line: data,
        2022-08-20 00:00:00            NaN
        2022-08-20 00:01:00            NaN
        2022-08-20 00:02:00            NaN
        2022-08-20 00:03:00            NaN
        2022-08-20 00:04:00            NaN
                                  ...
        2022-08-20 23:55:00    1568.069688
        2022-08-20 23:56:00    1568.036998
        2022-08-20 23:57:00    1568.004837
        2022-08-20 23:58:00    1567.990103
        2022-08-20 23:59:00    1567.975368
        Freq: T, Name: price, Length: 1440, dtype: float64
        :type line: Line
        """
        if not isinstance(line.index, pd.core.indexes.datetimes.DatetimeIndex):
            raise DemeterError("date index must be datetime")
        if isinstance(market, MarketInfo):
            self.broker.markets[market].data[name] = line
        elif isinstance(market, Market):
            market.data[name] = line
        else:
            raise DemeterError(f"{market} is not a valid market")



import logging
import os
import pickle
import time
from datetime import datetime
from typing import List, Dict

import orjson
import pandas as pd
from tqdm import tqdm  # process bar

from .evaluating_indicator import Evaluator
from .. import Broker, RowData, Asset
from .._typing import DemeterError, EvaluatorEnum, UnitDecimal, PositionInfo
from ..broker import BaseAction, AccountStatus, MarketInfo, MarketDict
from ..uniswap import UniLpMarket
from ..strategy import Strategy
from ..utils import get_formatted_predefined, STYLE


class Actuator(object):
    """
    Core component of a back test. Manage the resources in a test, including broker/strategy/data/indicator,



    """

    def __init__(self, allow_negative_balance=False):
        """
        init Actuator
        :param allow_negative_balance: balance can less than 0
        """
        # all the actions during the test(buy/sell/add liquidity)
        self._action_list: List[BaseAction] = []
        self._currents = {
            "actions": [],
            "timestamp": None
        }
        # broker status in every bar, use array for performance
        self._account_status_list: List[AccountStatus] = []
        self._account_status_df: pd.DataFrame = None

        # broker
        self._broker: Broker = Broker(allow_negative_balance, self._record_action_list)
        # strategy
        self._strategy: Strategy = Strategy()
        self._token_prices: pd.DataFrame = None
        # path of source data, which is saved by downloader
        # evaluating indicator calculator
        self._evaluator: Evaluator = None
        self._enabled_evaluator: [] = []
        # logging
        logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
        self.logger = logging.getLogger(__name__)

        # internal var
        self.__backtest_finished = False

    def _record_action_list(self, action):
        """
        record action list
        :param action: action
        :return: None
        """
        action.timestamp = self._currents["timestamp"]
        action.set_type()
        self._action_list.append(action)
        self._currents["actions"].append(action)

    # region property
    @property
    def account_status(self) -> List[AccountStatus]:
        """
        account status of all market,
        """
        return self._account_status_list

    @property
    def token_prices(self):
        """
        price of all token
        :return: None
        """
        return self._token_prices

    @property
    def final_status(self) -> AccountStatus:
        """
        Get status after back test finish.

        If test has not run, an error will be raised.

        :return: Final state of broker
        :rtype: AccountStatus
        """
        if self.__backtest_finished:
            return self._account_status_list[len(self._account_status_list) - 1]
        else:
            raise DemeterError("please run strategy first")

    def reset(self):
        """

        reset all the status variables

        """
        self._evaluator: Evaluator = None
        self._enabled_evaluator: [] = []

        self._action_list = []
        self._currents = {"actions": [], "timestamp": None}
        self._account_status_list = []
        self.__backtest_finished = False

        self._account_status_df: pd.DataFrame = None

    @property
    def actions(self) -> [BaseAction]:
        """
        all the actions during the test(buy/sell/add liquidity)

        :return: action list
        :rtype: [BaseAction]
        """
        return self._action_list

    @property
    def evaluating_indicator(self) -> Dict[EvaluatorEnum, UnitDecimal]:
        """
        evaluating indicator result

        :return:  evaluating indicator
        :rtype: EvaluatingIndicator
        """
        return self._evaluator.result if self._evaluator is not None else None

    @property
    def broker(self) -> Broker:
        """
        Broker manage assets in back testing. Including asset, positions. it also provides operations for positions,


        """
        return self._broker

    @property
    def strategy(self) -> Strategy:
        """
        strategy,

        :return: strategy
        :rtype: Strategy
        """
        return self._strategy

    @strategy.setter
    def strategy(self, value):
        """
        set strategy
        :param value: strategy
        :type value: Strategy
        """
        if isinstance(value, Strategy):
            self._strategy = value
        else:
            raise ValueError()

    @property
    def number_format(self) -> str:
        """
        number format for console output, eg: ".8g", ".5f"

        :return: number format
        :rtype: str
        """
        return self._number_format

    @number_format.setter
    def number_format(self, value: str):
        """
        number format for console output, eg: ".8g", ".5f",
        follow the document here: https://python-reference.readthedocs.io/en/latest/docs/functions/format.html

        :param value: number format,
        :type value:str
        """
        self._number_format = value

    # endregion

    def get_account_status_dataframe(self) -> pd.DataFrame:
        """
        get account status dataframe
        :return: dataframe
        """
        return AccountStatus.to_dataframe(self._account_status_list)

    def set_assets(self, assets: List[Asset]):
        """
        set initial balance for token

        :param assets: assets to set.
        :type assets: [Asset]
        """
        for asset in assets:
            self._broker.set_balance(asset.token_info, asset.balance)

    def set_price(self, prices: pd.DataFrame | pd.Series):
        """
        set price
        :param prices: dataframe or series
                                                       eth  usdc
        2022-08-20 00:00:00  1610.553895752868641174609110     1
        2022-08-20 00:01:00  1612.487623747744872677867817     1
        2022-08-20 00:02:00  1615.715664560742874527210287     1
        2022-08-20 00:03:00  1615.715664560742874527210287     1
        2022-08-20 00:04:00  1615.554109149827891738036484     1
        ...                                            ...   ...
        2022-08-20 23:55:00  1577.086574012079067849553855     1
        2022-08-20 23:56:00  1576.928881123966671182435611     1
        2022-08-20 23:57:00  1576.928881123966671182435611     1
        2022-08-20 23:58:00  1576.613542649301384412539259     1
        2022-08-20 23:59:00  1576.613542649301384412539259     1
        [1440 rows x 2 columns]
        :return: None
        """
        if isinstance(prices, pd.DataFrame):
            if self._token_prices is None:
                self._token_prices = prices
            else:
                self._token_prices = pd.concat([self._token_prices, prices])
        else:
            if self._token_prices is None:
                self._token_prices = pd.DataFrame(data=prices, index=prices.index)
            else:
                self._token_prices[prices.name] = prices

    def notify(self, strategy: Strategy, actions: List[BaseAction]):
        """

        notify user when new action happens.

        :param strategy: Strategy
        :type strategy: Strategy
        :param actions:  action list
        :type actions: [BaseAction]
        """
        if len(actions) < 1:
            return
        # last_time = datetime(1970, 1, 1)
        for action in actions:
            # if last_time != action.timestamp:
            #     print(f"\033[7;34m{action.timestamp} \033[0m")
            #     last_time = action.timestamp
            strategy.notify(action)

    def _check_backtest(self):
        """
        check backtest result
        :return:
        """
        # ensure a market exist
        if len(self._broker.markets) < 1:
            raise DemeterError("No market assigned")
        # ensure all token has price list.
        if self._token_prices is None:
            # if price is not set and market is uni_lp_market, get price from market automatically
            for market in self.broker.markets.values():
                if isinstance(market, UniLpMarket):
                    self.set_price(market.get_price_from_data())
            if self._token_prices is None:
                raise DemeterError("token prices is not set")
        for token in self._broker.assets.keys():  # dict_keys([TokenInfo(name='usdc', decimal=6), TokenInfo(name='eth', decimal=18)])
            if token.name not in self._token_prices:
                raise DemeterError(f"Price of {token.name} has not set yet")
        [market.check_before_test() for market in self._broker.markets.values()]

        data_length = []  # [1440]
        for market in self._broker.markets.values():
            data_length.append(len(market.data.index))
            market.check_asset()  # check each market, including assets
        # ensure data length same
        if List.count(data_length, data_length[0]) != len(data_length):
            raise DemeterError("data length among markets are not same")
        if len(self._token_prices.index) != data_length[0]:
            raise DemeterError("price length and data length are not same")
        length = data_length[0]
        # ensure data interval same
        data_interval = []
        if length > 1:
            for market in self._broker.markets.values():
                data_interval.append(market.data.index[1] - market.data.index[0])
            if List.count(data_interval, data_interval[0]) != len(data_interval):
                raise DemeterError("data interval among markets are not same")
            price_interval = self._token_prices.index[1] - self._token_prices.index[0]
            if price_interval != data_interval[0]:
                raise DemeterError("price list interval and data interval are not same")

    def __get_market_row_dict(self, index, row_id) -> MarketDict:
        """
        get market row dict info
        :param index:
        :param row_id:
        :return: Market dict
        """
        market_dict = MarketDict()
        for market_key, market in self._broker.markets.items():
            market_row = RowData(index.to_pydatetime(), row_id)
            df_row = market.data.loc[index]
            for column_name in df_row.index:
                setattr(market_row, column_name, df_row[column_name])
            market_dict[market_key] = market_row
        market_dict.set_default_key(self.broker.markets.get_default_key())
        return market_dict

    def __set_row_to_markets(self, timestamp, market_row_dict: dict):
        """
        set markets row data
        :param timestamp:
        :param market_row_dict:
        :return:
        """
        for market_key, market_row_data in market_row_dict.items():
            self._broker.markets[market_key].set_market_status(timestamp, market_row_data)

    def run(self,
            evaluator: List[EvaluatorEnum] = [],
            output: bool = True):
        """
        start back test, the whole process including:

        * reset actuator
        * initialize strategy (set object to strategy, then run strategy.initialize())
        * process each bar in data
            * prepare data in each row
            * run strategy.on_bar()
            * calculate fee earned
            * get latest account status
            * notify actions
        * run evaluator indicator
        * run strategy.finalize()

        :param enable_notify: notify when new action happens
        :type enable_notify: bool
        :param enable_evaluating: enable evaluating indicator. if not enabled, no evaluating will be calculated
        :type enable_evaluating: bool
        :param print_final_status: enable output.
        :type print_final_status: bool
        """
        run_begin_time = time.time()  # 1681718968.267463
        self.reset()

        self._enabled_evaluator = evaluator
        self._check_backtest()
        index_array: pd.DatetimeIndex = list(self._broker.markets.values())[0].data.index
        self.logger.info("init strategy...")

        # set initial status for strategy, so user can run some calculation in initial function.
        init_row_data = self.__get_market_row_dict(index_array[0], 0)
        self.__set_row_to_markets(index_array[0], init_row_data)
        # keep initial balance for evaluating
        init_account_status = self._broker.get_account_status(self._token_prices.iloc[0])
        self.init_strategy()
        row_id = 0
        data_length = len(index_array)
        first = True  # todo delete
        self.logger.info("start main loop...")
        with tqdm(total=data_length, ncols=150) as pbar:
            for timestamp_index in index_array:
                # prepare data of a row
                market_row_dict = self.__get_market_row_dict(timestamp_index, row_id)
                row_id += 1
                self.__set_row_to_markets(timestamp_index, market_row_dict)
                # execute strategy, and some calculate
                self._currents["timestamp"] = timestamp_index.to_pydatetime()
                self._strategy.before_bar(market_row_dict)

                if self._strategy.triggers:
                    for trigger in self._strategy.triggers:
                        if trigger.when(market_row_dict):
                            trigger.do(market_row_dict)
                self._strategy.on_bar(market_row_dict)

                # update broker status, eg: re-calculate fee
                # and read the latest status from broker
                for market in self._broker.markets.values():
                    market.update()

                self._strategy.after_bar(market_row_dict)

                if first:  # todo delete
                    first = False  # todo delete
                self._account_status_list.append(
                    self._broker.get_account_status(self._token_prices.loc[timestamp_index],
                                                    timestamp_index))
                # notify actions in current loop
                self.notify(self.strategy, self._currents["actions"])
                self._currents["actions"] = []
                # move forward for process bar and index
                pbar.update()

        self.logger.info("main loop finished")
        self._account_status_df: pd.DataFrame = self.get_account_status_dataframe()

        if len(self._enabled_evaluator) > 0:
            self.logger.info("Start calculate evaluating indicator...")
            self._evaluator = Evaluator(init_account_status, self._account_status_df, self._token_prices)
            self._evaluator.run(self._enabled_evaluator)
            self.logger.info("Evaluating indicator has finished it's job.")
        self._strategy.finalize()
        self.__backtest_finished = True
        if output:
            self.output()

        self.logger.info(f"Backtesting finished, execute time {time.time() - run_begin_time}s")

    def output(self):
        """
        output back test result to console
        """
        if not self.__backtest_finished:
            raise DemeterError("Please run strategy first")
        self.logger.info(self.broker.formatted_str())
        self.logger.info(get_formatted_predefined("Account Status", STYLE["header1"]))
        self.logger.info(self._account_status_df)
        if len(self._enabled_evaluator) > 0:
            self.logger.info("Evaluating indicator")
            self.logger.info(self._evaluator)

    def save_result(self, path: str, account=True, actions=True) -> List[str]:
        """
        save back test result
        :param path: path to save
        :type path: str
        :param account: Save account status or not
        :type account: bool
        :param actions: Save actions or not
        :type actions: bool
        :return:
        :rtype:
        """
        if not self.__backtest_finished:
            raise DemeterError("Please run strategy first")
        file_name_head = "backtest-" + datetime.now().strftime('%Y%m%d-%H%M%S')
        if not os.path.exists(path):
            os.mkdir(path)
        file_list = []
        if account:
            file_name = os.path.join(path, file_name_head + ".account.csv")
            self._account_status_df.to_csv(file_name)
            file_list.append(file_name)
        if actions:
            # save pkl file to load again
            pkl_name = os.path.join(path, file_name_head + ".action.pkl")
            with open(pkl_name, "wb") as outfile1:
                pickle.dump(self._action_list, outfile1)
            # save json to read
            actions_json_str = orjson.dumps(self._action_list,
                                            option=orjson.OPT_INDENT_2,
                                            default=json_default)
            json_name = os.path.join(path, file_name_head + ".action.json")
            with open(json_name, "wb") as outfile:
                outfile.write(actions_json_str)

            file_list.append(json_name)
            file_list.append(pkl_name)

        self.logger.info("files have saved to", file_list)
        return file_list

    def init_strategy(self):
        """
        initialize strategy, set property to strategy. and run strategy.initialize()
        """
        if not isinstance(self._strategy, Strategy):
            raise DemeterError("strategy must be inherit from Strategy")
        self._strategy.broker = self._broker
        self._strategy.markets = self._broker.markets
        market_datas = MarketDict()
        for k, v in self.broker.markets.items():
            market_datas[k] = v.data
        market_datas.set_default_key(self.broker.markets.get_default_key())
        self._strategy.data = market_datas
        self._strategy.prices = self._token_prices
        self._strategy.account_status = self._account_status_list
        self._strategy.actions = self._action_list
        self._strategy.assets = self.broker.assets
        self._strategy.get_account_status_dataframe = self.get_account_status_dataframe
        for k, v in self.broker.markets.items():
            setattr(self._strategy, k.name, v)
        for k, v in self.broker.assets.items():
            setattr(self._strategy, k.name, v)
        self._strategy.initialize()

    def __str__(self):
        return f"Demeter Actuator (broker:{self._broker})\n"


def json_default(obj):
    """
    format json data
    :param obj:
    :return:
    """
    if isinstance(obj, UnitDecimal):
        return obj.to_str()
    elif isinstance(obj, MarketInfo):
        return {"name": obj.name}
    elif isinstance(obj, PositionInfo):
        return {"lower_tick": obj.lower_tick,
                "upper_tick": obj.upper_tick}
    else:
        raise TypeError



from ._typing import DemeterError, EthError, TokenInfo, UnitDecimal, DECIMAL_0, DECIMAL_1, EvaluatorEnum
from .broker import Broker, RowData, MarketInfo, Asset, MarketDict, AssetDict, AccountStatus
from .core import Actuator
from .download import ChainType, DataSource
from .indicator import simple_moving_average, exponential_moving_average, actual_volatility
from .strategy import Strategy, Trigger, TimeRangesTrigger, TimeRangeTrigger, TimeRange, PeriodTrigger, \
    PeriodsTrigger, AtTimesTrigger, AtTimeTrigger
from .uniswap import UniLpMarket, UniV3Pool, UniV3PoolStatus, BrokerAsset, Position



from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import Dict

import pandas as pd
from pandas import _typing as pd_typing

from ..broker import Rule, RowData

DEFAULT_AGG_METHOD = "first"
EMPTY_RULE = Rule(None, None, None)


@dataclass
class UniLPDataRaw:
    """
    data types in csv file saved by download module
    """
    timestamp: datetime = None
    netAmount0: int = None
    netAmount1: int = None
    closeTick: int = None
    openTick: int = None
    lowestTick: int = None
    highestTick: int = None
    inAmount0: int = None
    inAmount1: int = None
    currentLiquidity: int = None


@dataclass
class UniLPData(RowData):
    """
    data type used in back test, extended from  UniLPDataRaw
    """
    netAmount0: int = None
    netAmount1: int = None
    closeTick: int = None
    openTick: int = None
    lowestTick: int = None
    highestTick: int = None
    inAmount0: int = None
    inAmount1: int = None
    currentLiquidity: int = None
    open: Decimal = None
    price: Decimal = None
    low: Decimal = None
    high: Decimal = None
    volume0: Decimal = None
    volume1: Decimal = None


class LineTypeEnum(Enum):
    """
    predefined column, used to define fillna method.
    """
    timestamp = 1
    netAmount0 = 2
    netAmount1 = 3
    closeTick = 4
    openTick = 5
    lowestTick = 6
    highestTick = 7
    inAmount0 = 8
    inAmount1 = 9
    currentLiquidity = 10
    other = 100


LINE_RULES = {
    LineTypeEnum.timestamp.name: EMPTY_RULE,
    LineTypeEnum.netAmount0.name: Rule("sum", None, 0),
    LineTypeEnum.netAmount1.name: Rule("sum", None, 0),
    LineTypeEnum.closeTick.name: Rule("last", "ffill", None),
    LineTypeEnum.openTick.name: Rule("first", "ffill", None),
    LineTypeEnum.lowestTick.name: Rule("min", "ffill", None),
    LineTypeEnum.highestTick.name: Rule("max", "ffill", None),
    LineTypeEnum.inAmount0.name: Rule("sum", None, 0),
    LineTypeEnum.inAmount1.name: Rule("sum", None, 0),
    LineTypeEnum.currentLiquidity.name: Rule("sum", "ffill", None),
}


def get_line_rules_safe(key: str) -> Rule:
    if key in LINE_RULES:
        return LINE_RULES[key]
    else:
        return EMPTY_RULE


def resample(df: pd.DataFrame,
             rule,
             axis=0,
             closed: str | None = None,
             label: str | None = None,
             convention: str = "start",
             kind: str | None = None,
             loffset=None,
             base: int | None = None,
             on=None,
             level=None,
             origin: str | pd_typing.TimestampConvertibleTypes = "start_day",
             offset: pd_typing.TimedeltaConvertibleTypes | None = None,
             agg: Dict[str, str] = None) -> pd.DataFrame:
    """
    resample data
    :param df: data in dataframe
    :param rule: resample rule, see Dataframe.resample doc
    :param axis: resample axis, see Dataframe.resample doc
    :param closed: resample closed, see Dataframe.resample doc
    :param label: resample label, see Dataframe.resample doc
    :param convention: resample convention, see Dataframe.resample doc
    :param kind: resample kind, see Dataframe.resample doc
    :param loffset: resample loffset, see Dataframe.resample doc
    :param base: resample base, see Dataframe.resample doc
    :param on: resample on, see Dataframe.resample doc
    :param level: resample level, see Dataframe.resample doc
    :param origin: resample origin, see Dataframe.resample doc
    :param offset: resample offset, see Dataframe.resample doc
    :param agg: aggregate method
    :return: aggregated dataframe
    """
    agg = agg if agg else {}
    resampler = df.resample(rule, axis, closed, label, convention, kind, loffset, base, on, level, origin, offset)
    agg_dict = {}
    for column_name in df.columns:
        rule = get_line_rules_safe(column_name)
        agg_method = rule.agg
        if agg_method is None:
            if column_name in agg:
                agg_method = agg[column_name]
            else:
                agg_method = DEFAULT_AGG_METHOD
        agg_dict[column_name] = agg_method
    df_new = resampler.agg(agg_dict)
    return df_new


def fillna(
        df: pd.DataFrame,
        value: object | pd_typing.ArrayLike | None = None,
        method: pd_typing.FillnaOptions | None = None,
        axis: pd_typing.Axis | None = None,
        inplace: bool = False,
        limit=None,
        downcast=None) -> pd.DataFrame | None:
    """
    fill empty item. param is the same to pandas.Series.fillna

    if column name is predefined, method and value will be omitted, and data will be filled as predefined

    """
    new_df = df.copy(False)

    # fill close tick first, it will be used later.
    if LineTypeEnum.closeTick.name in new_df.columns:
        new_df[LineTypeEnum.closeTick.name] = \
            new_df[LineTypeEnum.closeTick.name].fillna(value=None,
                                                       method=get_line_rules_safe(
                                                           LineTypeEnum.closeTick.name).fillna_method,
                                                       axis=axis,
                                                       inplace=inplace,
                                                       limit=limit,
                                                       downcast=downcast)
    for column_name in new_df.columns:
        if column_name == LineTypeEnum.closeTick.name:
            continue
        rule = get_line_rules_safe(column_name)
        if not rule.fillna_method and rule.fillna_value is None:
            new_df[column_name] = new_df[column_name].fillna(value, method, axis, inplace, limit, downcast)
        else:
            current_method = rule.fillna_method if rule.fillna_method else method
            current_value = rule.fillna_value if rule.fillna_value is not None else value
            # all tick related field will be filled with close_tick.
            if column_name in [LineTypeEnum.openTick.name,
                               LineTypeEnum.highestTick.name,
                               LineTypeEnum.lowestTick.name] and LineTypeEnum.closeTick.name in new_df.columns:
                current_method = None
                current_value = new_df[LineTypeEnum.closeTick.name]
            new_df[column_name] = new_df[column_name].fillna(value=current_value, method=current_method, axis=axis,
                                                             inplace=inplace,
                                                             limit=limit,
                                                             downcast=downcast)
    return new_df
